# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

Modified By: hsky77
Last Updated: November 22nd 2020 20:41:28 pm
'''


import os
import re
import sys
from typing import List, Set, ByteString, Tuple

from hyssop.util import walk_to_file_paths, join_path, configure_colored_logging

from .constants import Dependency_Folder, Project_Pack_File, Project_Requirement_File


class HyssopPack(object):
    """
    pack server files to compressed output file, .zip for windows os and tar.gz for the others
    """

    default_exclude_keys = ['__pycache__']

    def pack(self, server_folder: str, output_file_path: str, prepare_wheels: bool = False,
             compile_py: bool = True, log_level: str = 'INFO'):
        configure_colored_logging(log_level)

        if output_file_path is None:
            file_name = os.path.basename(os.path.normpath(server_folder))
            if sys.platform == 'win32':
                output_file_path = 'win_' + file_name + '.zip'
            else:
                output_file_path = 'linux_' + file_name + '.tar.gz'

        # create packing file paths
        paths, py_paths, requirements = self._create_file_paths(server_folder)

        # get wheel dependencies
        if prepare_wheels:
            for req in requirements:
                if os.path.isfile(req):
                    self._preparing_python_wheels(Dependency_Folder, req)
            paths.update(walk_to_file_paths(Dependency_Folder))

        # add arc path for iterable paths as [(path, arcpath)]
        paths = self._setup_arc_paths(paths, server_folder)
        py_paths = self._setup_arc_paths(py_paths, server_folder)
        requirements = self._setup_arc_paths(requirements, server_folder)

        # making ouput file
        self._packing_files(
            paths, py_paths, output_file_path, compile_py=compile_py)

    def _setup_arc_paths(self, paths: str, server_folder: str):
        new_paths = []
        server_folder = os.path.dirname(server_folder)
        for path in paths:
            if os.path.isabs(path):
                arc_path = join_path(os.path.relpath(path, server_folder))
            else:
                arc_path = join_path(os.path.relpath(
                    join_path(server_folder, path), server_folder))
            new_paths.append((path, arc_path))
        return new_paths

    def _create_file_paths(self, server_folder: str) -> Tuple[Set[str], Set[str], Set[str]]:
        import yaml
        print('creating packing list...')
        paths = set()
        py_paths = set()
        requirement_paths = set()

        hyssop_exclude = self.default_exclude_keys
        for directory in [server_folder]:
            sub_paths = set()
            pack_list = join_path(directory, Project_Pack_File)
            if os.path.isfile(pack_list):
                with open(pack_list, 'r') as f:
                    config = yaml.load(f, Loader=yaml.SafeLoader)
                    config = config if config else {}

                    if 'include' in config:
                        if config['include'] is not None:
                            for path in config['include']:
                                if os.path.isdir(path):
                                    ps, pys, rs = self._create_file_paths(path)
                                    paths.update(ps)
                                    py_paths.update(pys)
                                    requirement_paths.update(rs)
                                else:
                                    sub_paths.add(path)

                    sub_paths.update(walk_to_file_paths(directory))

                    if 'exclude' in config:
                        if config['exclude'] is not None:
                            exclude = hyssop_exclude + config['exclude']
                        else:
                            exclude = hyssop_exclude
                    exclude.append(pack_list)

                    py_paths.update([p for p in sub_paths if '.py' in p and max(
                        [p.find(x) for x in exclude]) == -1])
                    paths.update([p for p in sub_paths if not '.py' in p and max(
                        [p.find(x) for x in exclude]) == -1])
            else:
                # if pack setting file does not exist, collect all files under the server folder
                sub_paths = walk_to_file_paths(directory)
                paths.update([p for p in sub_paths if not '.py' in p and max(
                    [p.find(x) for x in hyssop_exclude]) == -1])
                py_paths.update([p for p in sub_paths if '.py' in p and max(
                    [p.find(x) for x in hyssop_exclude]) == -1])

            requirement_paths.add(
                join_path(directory, Project_Requirement_File))

        return paths, py_paths, requirement_paths

    def _preparing_python_wheels(self, dependency_folder: str, requirements: str) -> None:
        """using pip to download python packages for later packing"""
        import subprocess

        print('clean dependency folder')
        self.__remove_dir(dependency_folder)

        # build wheels
        if sys.platform == 'win32':
            command = 'python'
        else:
            command = 'python3'
        print('preparing pip wheels...')
        res = subprocess.run([command, '-m', 'pip', 'wheel', '-w',
                              dependency_folder, '-r', requirements])

        if res.returncode > 0:
            raise Exception('getting pip wheels failed...')

    def _packing_files(self, paths: List[str], py_paths: List[str], output_path: str, compile_py: bool = True) -> None:
        if len(paths) > 0:
            if sys.platform == 'win32':
                import zipfile
                with zipfile.ZipFile(output_path, mode='w') as zf:
                    for path in paths:
                        print('packing file: {}'.format(path[0]))
                        zf.write(path[0], arcname=path[1])

                    for path in py_paths:
                        if compile_py:
                            bytecode = self.__get_compiled_py_bytecode(path[0])
                            zf.writestr(path[1].replace(
                                '.py', '.pyc'), bytecode)
                        else:
                            print('packing py file: {}'.format(path[0]))
                            zf.write(path[0], arcname=path[1])
            else:
                import tarfile
                from io import BytesIO
                with tarfile.open(output_path, mode='w:gz') as tarf:
                    for path in paths:
                        print('packing file: {}'.format(path))
                        tarf.add(path[0], arcname=path[1])

                    for path in py_paths:
                        if compile_py:
                            bytecode = self.__get_compiled_py_bytecode(path[0])
                            bIO = BytesIO()
                            bIO.write(bytecode)
                            bIO.seek(0)
                            info = tarfile.TarInfo(
                                name=path[1].replace('.py', '.pyc'))
                            info.size = len(bIO.getbuffer())
                            tarf.addfile(tarinfo=info, fileobj=bIO)
                        else:
                            print('packing py file: {}'.format(path[0]))
                            tarf.add(path[0], arcname=path[1])
        else:
            print('Warning: no file had been packed to {}'.format(output_path))

    def __remove_dir(self, dir: str) -> None:
        from shutil import rmtree
        try:
            rmtree(dir)
        except:
            pass

    def __get_compiled_py_bytecode(self, path: str) -> ByteString:
        import importlib
        import sys
        print('compiling and packing python file: {}'.format(path))
        loader = importlib.machinery.SourceFileLoader(
            '<hyssop_compiled>', path)
        source_bytes = loader.get_data(path)
        code = compile(source_bytes, path,
                       'exec', dont_inherit=True)
        source_stats = loader.path_stats(path)
        if sys.version_info.minor < 7:
            bytecode = importlib._bootstrap_external._code_to_bytecode(
                code, source_stats['mtime'], source_stats['size'])
        else:
            bytecode = importlib._bootstrap_external._code_to_timestamp_pyc(
                code, source_stats['mtime'], source_stats['size'])
        return bytecode



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: August 27th 2020 12:59:54 pm
'''


import os
from typing import List, Dict
from importlib import import_module


def get_class(module: str, *attrs):
    """cls = get_class("module", "class / static function", "class static function")"""
    cls = import_module(module)
    for a in attrs:
        cls = getattr(cls, a)
    return cls


def join_to_abs_path(*paths) -> str:
    """os.path.join() + os.path.abspath() to return only linux path string, that means replace backslash with slash"""
    return os.path.abspath(os.path.join(*paths)).replace('\\', '/')


def join_path(*paths) -> str:
    """os.path.join() to return only linux path string, that means replace backslash with slash"""
    return os.path.join(*paths).replace('\\', '/')


def walk_to_file_paths(file_or_directory: str) -> List[str]:
    """get a list of absolutely path from the input path recursively"""
    file_paths = []
    if os.path.isdir(file_or_directory):
        for root, _, files in os.walk(file_or_directory):
            for i in range(len(files)):
                files[i] = join_path(os.path.abspath(root), files[i])

            if len(files) > 0:
                file_paths.extend(files)
    elif os.path.isfile(file_or_directory):
        file_paths.append(file_or_directory)

    return file_paths



# Copyright (C) 2019-Present the hyssop authors and contributors
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php:

'''
server loads all default components in hyssop.project.component.DefaultComponentTypes when starting

    LocalizationComponent:

        - managing localized message via config setting such as:

        component:
            localization:
                dir: 'local'    # optional: Load the .csv files in the directory under project directory.
                lang: 'en'      # optional: Setup language, default: en.

    LoggerComponent:

        - managing logger via config setting such as:

        component:
            logger:
                log_to_resources: False     # optional: Enable log to resources
                log_to_console: False       # optional: Enable log to console
                dir: 'logs'                 # optional. Log to files in the folder under porject directory if specified

    CallbackComponent:

        - managing callback function by enum types, no configuration settings needed

    ExecutorComponent:

        - managing executor to execute the given functions and arguments

        component:
            executor:
                worker_count: 1      # The maximum of workers is 2

Last Updated: November 22nd 2020 14:48:31 pm
'''

import logging

from datetime import datetime
from enum import Enum
from typing import Callable, Any, List, Dict, ByteString

from . import Component, ComponentManager, ComponentTypes
from ...util import ExecutorFactory, Executor, Callbacks, configure_colored_logging, join_path, BaseSyncLogger
from .mixin import FileLoggerMixin


class LocalizationComponent(Component):
    """default component for managing localized message by config setting"""

    def init(self, component_manager: ComponentManager, lang: str = 'en', **kwargs) -> None:
        import os
        from hyssop.util import BaseLocal
        self.local = BaseLocal

        self.dir = kwargs.get('dir', None)
        if self.dir is not None:
            self.dir = join_path(kwargs.get('project_dir', ''), self.dir)
            self.local.import_csvs_from_directory(self.dir)

        if lang is not None:
            self.local.set_language(lang)

    def info(self) -> str:
        return {**super().info(), **{'info': self.local.get_info()}}

    @property
    def current_language(self) -> str:
        return self.local.current_language

    def set_language(self, lang: str) -> None:
        self.local.set_language(lang)

    def get_message(self, code: str, *args) -> str:
        """convert to localized message with code and following parameters"""
        return self.local.get_message(code, *args)


class LoggerComponent(FileLoggerMixin, Component):
    """default component for managing logger by server config"""

    default_loggers = []

    def __init__(self, component_type: ComponentTypes):
        super().__init__(component_type)
        configure_colored_logging()
        self.loggers = {}
        self.kwargs = None

    def init(self, component_manager: ComponentManager, **kwargs) -> None:
        self.kwargs = kwargs
        self.log_level = logging.INFO
        self.log_echo = self.kwargs.get('log_to_console', False)
        self.log_to_resources = self.kwargs.get('log_to_resources', False)

    def get_logger(self, name: str, *args, sub_dir: str = '', mode: str = 'a', encoding: str = 'utf-8', echo: bool = False) -> BaseSyncLogger:
        """create and return logger object, sub_dir appends the path to configured log path"""
        logger = self.loggers[name] if name in self.loggers else None

        if not logger:
            logger = logging.getLogger(name)

            if self.log_to_resources:
                self.update_file_handler(
                    logger, sub_dir, mode, encoding, **self.kwargs)
            else:
                self.remove_file_handler(logger, sub_dir, **self.kwargs)

            self.loggers[name] = logger

        if logger:
            logger.setLevel(self.log_level)
            logger.propagate = self.log_echo or echo

        return logger

    def update_default_logger(self, debug: bool = False) -> None:
        self.log_level = logging.DEBUG if debug else self.log_level

        for name in self.default_loggers:
            self.get_logger(name)

        self.loggers.clear()


class CallbackComponent(Component):
    """default component for managing callback function by enum types"""

    def init(self, component_manager: ComponentManager, *arugs, **kwargs) -> None:
        self._callback_manager = {}

    def get_callback_obj(self, enum_cls: Enum) -> Callbacks:
        if not enum_cls in self._callback_manager:
            self._callback_manager[enum_cls] = Callbacks(enum_cls)
        return self._callback_manager[enum_cls]

    def add_callback(self, callback_enum_type: Enum, callback: Callable) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if not enum_cls in self._callback_manager:
                self._callback_manager[enum_cls] = Callbacks(enum_cls)

            self._callback_manager[enum_cls].add_callback(
                callback_enum_type, callback)

    def remove_callback(self, callback_enum_type: Enum, callback: Callable) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                self._callback_manager[enum_cls].remove_callback(
                    callback_enum_type, callback)

    def execute_callback(self, callback_enum_type: Enum, *args, **kwargs) -> None:
        from tornado.ioloop import IOLoop
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                self._callback_manager[enum_cls].execute_callback(
                    callback_enum_type, *args, **kwargs)

    async def execute_callback_async(self, callback_enum_type: Enum, *args, **kwargs) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                await self._callback_manager[enum_cls].execute_callback_async(
                    callback_enum_type, *args, **kwargs)


class ExecutorComponent(Component):
    def init(self, component_manager: ComponentManager, worker_count: int = 1, **kwargs) -> None:
        self.executor_factory = ExecutorFactory(worker_limit=worker_count)
        self.disposing = False

    def info(self) -> Dict:
        return {**super().info(), **{
            'info': {
                'workers': self.executor_factory.worker_count
            }
        }}

    def run_method_in_queue(self,
                            func: Callable,
                            *args,
                            on_finish: Callable[[Any], None] = None,
                            on_exception: Callable[[Exception], None] = None,
                            **kwargs) -> None:
        if not self.disposing:
            self.executor_factory.run_method_in_queue(
                func, *args, on_finish=on_finish, on_exception=on_exception, **kwargs)

    def run_method(self, func: Callable, *args, **kwargs) -> Any:
        if not self.disposing:
            return self.executor_factory.run_method(func, *args, **kwargs)

    async def run_method_async(self, func: Callable, *args, **kwargs) -> Any:
        if not self.disposing:
            return await self.executor_factory.run_method_async(func, *args, **kwargs)

    def get_executor(self) -> Executor:
        return self.executor_factory.get_executor()

    def dispose(self, component_manager: ComponentManager) -> None:
        self.disposing = True
        self.executor_factory.dispose()



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: August 27th 2020 12:55:07 pm
'''


from typing import Dict, List, Any, Tuple

from .localization import BaseLocal
from .constants import LocalCode_Not_HierarchyElementMeta_Subclass, LocalCode_No_Parameters, LocalCode_Parameters_No_Key


class HierarchyElementMeta(type):
    def set_cls_parameters(cls, *cls_parameters) -> None:
        for cls_parameter in cls_parameters:
            if not hasattr(cls, '_cls_parameters'):
                cls._cls_parameters = {}

            if issubclass(type(cls), HierarchyElementMeta):
                cls._cls_parameters[cls_parameter.__name__] = cls_parameter
            else:
                raise TypeError(BaseLocal.get_message(
                    LocalCode_Not_HierarchyElementMeta_Subclass, cls))

    def get_cls_parameter(cls, key_routes: str, delimeter='.') -> type:
        routes = key_routes.split(delimeter)
        temp = cls
        for key in routes:
            if hasattr(temp, '_cls_parameters'):
                if key in temp._cls_parameters:
                    temp = temp._cls_parameters[key]
                else:
                    raise KeyError(BaseLocal.get_message(
                        LocalCode_Parameters_No_Key, temp, key))
            else:
                raise AttributeError(BaseLocal.get_message(
                    LocalCode_No_Parameters, temp))

        return temp



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:35:38 pm
'''

from typing import Dict, List, Type, Any, Tuple, Union, Iterable

from enum import Enum

from ..util import HierarchyElementMeta, LocalCode_No_Parameters, LocalCode_Parameters_No_Key, BaseLocal

from .constants import (LocalCode_Parameter_Required, LocalCode_Parameter_Type_Error, LocalCode_Setup_Error,
                        LocalCode_Invalid_Parameter, LocalCode_Copied_Name_Equal_To_Origin)


class ConfigElementType(Enum):
    Container = 'container'
    Element = 'element'
    ScalableContainer = 'scalable_container'
    ScalableElement = 'scalable_element'
    SwitchableElement = 'switchable_element'


class ConfigBaseElementMeta(HierarchyElementMeta):
    def __new__(cls, name: str, parameter_type: Any, required: bool = True, *cls_parameters) -> type:
        return super().__new__(cls, name, (object,), {
            'parameter_type': parameter_type, 'required': required, 'get_parameter': cls.get_parameter})

    def __init__(cls, name: str, parameter_type: Any, required: bool, *cls_parameters) -> None:
        super().__init__(name, (object,), {})
        cls.set_cls_parameters(*cls_parameters)

    def __call__(cls, config: Union[Dict, List, str, int, float, bool], logs: List[str] = None, route: str = ''):
        obj = super().__call__()
        try:
            if config is not None:
                obj.parameter_type(config)
            obj.parameter = config
            obj.config_route = route
            obj._cls = cls
            if isinstance(logs, list) and not isinstance(config, dict):
                logs.append('{}: {}'.format(route, config))
        except:
            raise TypeError(BaseLocal.get_message(
                LocalCode_Parameter_Type_Error, route, cls.__name__, config, obj.parameter_type.__name__))

        if hasattr(cls, '_cls_parameters'):
            if config is not None:
                ConfigBaseElementMeta._convert_scalable_elements(cls, config)

            for k, v in cls._cls_parameters.items():
                if v.required:
                    if config is None or not k in config:
                        raise KeyError(BaseLocal.get_message(
                            LocalCode_Parameter_Required, cls.__name__, k))

                if config is not None and k in config:
                    if not hasattr(obj, '_parameters'):
                        obj._parameters = {}
                    obj._parameters[k] = v(config[k], logs, '{}.{}'.format(
                        route, k) if not route == '' else k)
        return obj

    def get_parameter(self, key_routes: str, delimeter: str = '.'):
        temp = self
        if key_routes is not None:
            routes = key_routes.split(delimeter)
            for key in routes:
                if hasattr(temp, '_parameters'):
                    if key in temp._parameters:
                        temp = temp._parameters[key]
                    else:
                        raise KeyError(BaseLocal.get_message(
                            LocalCode_Parameters_No_Key, temp, key))
                else:
                    raise KeyError(BaseLocal.get_message(
                        LocalCode_No_Parameters, temp))

        return temp.parameter

    def _convert_scalable_elements(cls, config: Dict):
        pops = []
        extends = []

        for k in cls._cls_parameters:  # convert scalable elements to identified elements
            if cls._cls_parameters[k].element_type is ConfigElementType.ScalableElement:
                names = [n for n in config if not n in cls._cls_parameters]
                for name in names:
                    extends.append(ConfigElementMeta(
                        name, cls._cls_parameters[k].parameter_type, cls._cls_parameters[k].required))
                pops.append(k)
            elif cls._cls_parameters[k].element_type is ConfigElementType.ScalableContainer:
                names = [cls._cls_parameters[k].parameter_type(
                    n) for n in config if not n in cls._cls_parameters]
                for name in names:
                    extends.append(ConfigContainerMeta(name, cls._cls_parameters[k].required,
                                                       *cls._cls_parameters[k]._cls_parameters.values()))
                pops.append(k)
            elif cls._cls_parameters[k].element_type is ConfigElementType.SwitchableElement:
                if config[k] in cls._cls_parameters[k]._cls_parameters:
                    for _, param_cls in cls._cls_parameters[k]._cls_parameters[config[k]]._cls_parameters.items():
                        extends.append(param_cls)
                    extends.append(
                        ConfigElementMeta(k, cls._cls_parameters[k].parameter_type, cls._cls_parameters[k].required))

                    pops.append(k)
                else:
                    raise RuntimeError(BaseLocal.get_message(
                        LocalCode_Setup_Error, cls))

        for k in pops:  # remove scalable or switchable element
            cls._cls_parameters.pop(k)

        for extend in extends:  # add scaled elements
            cls._cls_parameters[extend.__name__] = extend


class ConfigContainerMeta(ConfigBaseElementMeta):
    def __new__(cls, name: str, required: bool, *parameters) -> type:
        return super().__new__(cls, name, dict, required, *parameters)

    def __init__(cls, name: str, required: bool, *parameters) -> None:
        cls.element_type = ConfigElementType.Container
        super().__init__(name, dict, required, *parameters)

    def copy(cls, name):
        if name == cls.__name__:
            raise ValueError(BaseLocal.get_message(
                LocalCode_Copied_Name_Equal_To_Origin, name, cls.__name__))
        new_cls = ConfigContainerMeta(
            name, cls.required, *cls._cls_parameters.values())
        return new_cls


class ConfigElementMeta(ConfigBaseElementMeta):
    def __new__(cls, name: str, parameter_type: Any, required: bool) -> type:
        return super().__new__(cls, name, parameter_type, required)

    def __init__(cls, name: str, parameter_type: Any, required: bool) -> None:
        cls.element_type = ConfigElementType.Element
        super().__init__(name, parameter_type, required)

    def copy(cls, name):
        if name == cls.__name__:
            raise ValueError(BaseLocal.get_message(
                LocalCode_Copied_Name_Equal_To_Origin, name, cls.__name__))
        new_cls = ConfigElementMeta(name, cls.parameter_type, cls.required)
        return new_cls


class ConfigScalableContainerMeta(ConfigBaseElementMeta):
    def __new__(cls, parameter_type: Union[str, int], *parameters) -> type:
        return super().__new__(cls, 'ConfigScalableContainerMeta', parameter_type, False, *parameters)

    def __init__(cls, parameter_type: Union[str, int], *parameters) -> None:
        cls.element_type = ConfigElementType.ScalableContainer
        super().__init__('ConfigScalableContainerMeta', parameter_type, False, *parameters)

    def copy(cls):
        new_cls = ConfigScalableContainerMeta(
            cls.parameter_type, *cls._cls_parameters.values())
        return new_cls


class ConfigScalableElementMeta(ConfigBaseElementMeta):
    def __new__(cls, element_type: Union[str, int], parameter_type: Any) -> type:
        return super().__new__(cls, 'ConfigScalableElementMeta', element_type, False)

    def __init__(cls, element_type: Union[str, int], parameter_type: Any) -> None:
        cls.element_type = ConfigElementType.ScalableElement
        super().__init__('ConfigScalableElementMeta', parameter_type, False)

    def copy(cls):
        new_cls = ConfigScalableElementMeta(
            cls.element_type, cls.parameter_type)
        return new_cls


class ConfigSwitchableElementMeta(ConfigBaseElementMeta):
    def __new__(cls, name: str, parameter_type: Any, required: bool, *parameters) -> type:
        return super().__new__(cls, name, parameter_type, required, *parameters)

    def __init__(cls, name: str, parameter_type: Any, required: bool, *parameters) -> None:
        cls.element_type = ConfigElementType.SwitchableElement
        super().__init__(name, parameter_type, required, *parameters)

    def copy(cls, name):
        if name == cls.__name__:
            raise ValueError(BaseLocal.get_message(
                LocalCode_Copied_Name_Equal_To_Origin, name, cls.__name__))
        new_cls = ConfigSwitchableElementMeta(
            name, cls.parameter_type, cls.required, *cls._cls_parameters.values())
        return new_cls


ProjectConfigValidator = ConfigContainerMeta(
    'root', True,
    ConfigElementMeta('name', str, False),
    ConfigElementMeta('debug', bool, True)
)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:35:07 pm
'''

from ..constants import Component_Module_Folder
from ..config_validator import ConfigContainerMeta, ConfigElementMeta, ConfigScalableElementMeta, ProjectConfigValidator


ConfigComponentValidator = ConfigContainerMeta(
    Component_Module_Folder, False,
    ConfigContainerMeta(
        'localization', False,
        ConfigElementMeta('dir', str, False),
        ConfigElementMeta('lang', str, False)
    ),
    ConfigContainerMeta(
        'logger', False,
        ConfigElementMeta('dir', str, False)
    ),
    ConfigContainerMeta(
        'executor', False,
        ConfigElementMeta('worker_count', int, True)
    )
)

ProjectConfigValidator.set_cls_parameters(ConfigComponentValidator)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 27th 2020

This module contains the "yaml" configurable controller classes for hyssop application.

    - to create and setup controller api:

        0. the hierarchy of folders and files looks like:

            server_directory/
                server_config.yaml
                controller/
                    __init__.py
                    foo.py

        1. "__init__.py" defines the controllers:
            from hyssop.project.web.controller import ControllerType

            class ExControllerType(ControllerType):
                Foo = ('foo', 'foo', 'Foo')

        2. "foo.py" contains the controller class:

            from hyssop.project.web.controller.tornado import RequestController

            class Foo(RequestController):

                def initialize(self, p1, **kwds):
                    self.p1 = p1

                async def get(self):
                    self.write('Hello Workd: ', self.p1)

        3. setup component block of "server_config.yaml" to tell hyssop server load the extend components "Foo":

            controller:                 # block to setup controllers
                /foo:                   # api route
                    enum: foo           # tells to load foo controller class
                    params:
                        p1: xxxx        # parameter p1 of Foo.initialize()

Modified By: hsky77
Last Updated: November 22nd 2020 14:19:18 pm
'''

from typing import List, Dict

from ...util import join_path, join_to_abs_path, BaseLocal, DynamicTypeEnum
from ..constants import Controller_Module_Folder, LocalCode_Failed_To_Load_Controller
from .config import ConfigControllerValidator


class ControllerType(DynamicTypeEnum):
    """base abstract controller enum class"""
    @staticmethod
    def get_controller_enum_class() -> List[DynamicTypeEnum]:
        try:
            return DynamicTypeEnum.get_dynamic_class_enum_class(Controller_Module_Folder)
        except:
            pass


def _get_controller_enum(key: str, contoller_types: List[ControllerType]) -> DynamicTypeEnum:
    for contoller_type in contoller_types:
        try:
            return contoller_type(key)
        except:
            continue


def get_project_controllers(server_setting: Dict, server_dir: str = None) -> List:
    controllers = []
    if Controller_Module_Folder in server_setting and isinstance(server_setting[Controller_Module_Folder], dict):
        contoller_types = []
        try:
            contoller_types = ControllerType.get_controller_enum_class()
        except:
            pass

        # validate controller config
        ConfigControllerValidator(server_setting[Controller_Module_Folder])

        for path, v in server_setting[Controller_Module_Folder].items():
            if 'enum' in v:
                t = _get_controller_enum(v['enum'], contoller_types)
                if t is None:
                    raise ImportError(BaseLocal.get_message(
                        LocalCode_Failed_To_Load_Controller, server_dir, v['enum']))
                try:
                    cls_type = t.import_class()
                except:
                    cls_type = t.import_function()

                params = v['params'] if 'params' in v and v['params'] is not None else {}

                controllers.append((path, cls_type, params))

    return controllers



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:34:57 pm
'''

from ..constants import Controller_Module_Folder
from ..config_validator import ConfigContainerMeta, ConfigElementMeta, ConfigScalableContainerMeta, ProjectConfigValidator

ProjectConfigValidator.set_cls_parameters(
    ConfigElementMeta('port', int, False))
ProjectConfigValidator.set_cls_parameters(ConfigContainerMeta(
    'ssl', False,
    ConfigElementMeta('crt', str, True),
    ConfigElementMeta('key', str, True),
    ConfigElementMeta('ca', str, False)
))


ConfigControllerValidator = ConfigContainerMeta(
    Controller_Module_Folder, False,
    ConfigScalableContainerMeta(
        str,
        ConfigElementMeta('enum', str, True),
        ConfigContainerMeta('params', False)
    )
)

ProjectConfigValidator.set_cls_parameters(ConfigControllerValidator)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: March 6th 2021 17:37:26 pm
'''


Module_Path = __path__[0]
Version = '1.1.1'



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:43:10 pm
'''

import os

from ...util import join_path, DynamicTypeEnum, join_to_abs_path, BaseLocal
from ..component import (ComponentManager, ComponentTypes,
                         DefaultComponentTypes, create_component_manager)
from ..constants import Component_Module_Folder, LocalCode_File_Not_Found, Project_Config_File
from ..mixin import ProjectMixin
from .controller import get_project_controllers


class WebApplicationMinin(ProjectMixin):
    @property
    def project_controllers(self):
        if not hasattr(self, '_project_controllers'):
            self._project_controllers = get_project_controllers(
                self.project_config, self.project_dir)

        return self._project_controllers

    @property
    def project_ssl_context(self):
        if 'ssl' in self.project_config:
            if not os.path.isfile(join_path(self.project_config['ssl']['crt'])):
                raise FileNotFoundError(BaseLocal.get_message(
                    LocalCode_File_Not_Found, join_path(self.project_config['ssl']['crt'])))
            if not os.path.isfile(join_path(self.project_config['ssl']['key'])):
                raise FileNotFoundError(BaseLocal.get_message(
                    LocalCode_File_Not_Found, join_path(self.project_config['ssl']['key'])))

            import ssl
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(join_path(self.project_config['ssl']['crt']),
                                    join_path(self.project_config['ssl']['key']))
            if 'ca' in self.project_config['ssl']:
                if not os.path.isfile(join_path(self.project_config['ssl']['ca'])):
                    raise FileNotFoundError(BaseLocal.get_message(
                        LocalCode_File_Not_Found, join_path(self.project_config['ssl']['ca'])))
                ssl_ctx.load_verify_locations(
                    join_path(self.project_config['ssl']['ca']))
            return ssl_ctx



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: November 22nd 2020 14:43:18 pm
'''

import logging
from threading import Lock

LOG_FORMAT = '%(asctime)s %(levelname)-8s %(name)s  - %(message)s'


def configure_colored_logging(loglevel: str = 'INFO', logger: logging.Logger = None) -> None:
    """
    note: only call once at the beginning of program
    """
    import coloredlogs
    field_styles = coloredlogs.DEFAULT_FIELD_STYLES.copy()
    field_styles['asctime'] = {}
    level_styles = coloredlogs.DEFAULT_LEVEL_STYLES.copy()
    level_styles['debug'] = {}
    if logger is not None:
        coloredlogs.install(
            level=loglevel,
            use_chroot=False,
            fmt=LOG_FORMAT,
            level_styles=level_styles,
            field_styles=field_styles,
            logger=logger)
    else:
        coloredlogs.install(
            level=loglevel,
            use_chroot=False,
            fmt=LOG_FORMAT,
            level_styles=level_styles,
            field_styles=field_styles)


class BaseSyncLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.INFO):
        super().__init__(name, level)
        self.lock = Lock()

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False):
        with self.lock:
            super()._log(level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info)


logging.setLoggerClass(BaseSyncLogger)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: November 22nd 2020 12:21:26 pm
'''

from ..util import join_path, BaseLocal
from ..util.constants import Localization_File

BaseLocal.import_csv([join_path(__path__[0], Localization_File)])



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: August 27th 2020 12:26:00 pm
'''


from enum import Enum
from typing import Callable, Any, Union, Tuple
from threading import get_ident
from inspect import iscoroutinefunction

from . import BaseLocal
from .constants import LocalCode_Not_Valid_Enum, LocalCode_Not_ASYNC_FUNC, LocalCode_Invalid_Thread_Safe_Call


class Callbacks():
    """
    Enum managed callback functions. Basically, this is useful to communicate between the controller instances.
    """

    def __init__(self, callback_enum_cls: Enum, thread_safe: bool = True):
        self.callbacks = {}
        self.callback_type_cls = callback_enum_cls
        self.thread_safe = thread_safe

    def add_callback(self, callback_enum: Enum, callback: Callable) -> None:
        """
        Register callback function with the given enum
        """
        callback_type = self.callback_type_cls(callback_enum)

        if callable(callback):
            if not callback_type in self.callbacks:
                self.callbacks[callback_type] = []

            self.callbacks[callback_type].append(self.__make_item(callback))

    def remove_callback(self, callback_enum: Enum, callback: Callable) -> None:
        """
        Unregister callback function with the given enum
        """
        callback_type = self.callback_type_cls(callback_enum)

        if callback_type in self.callbacks:
            self.callbacks[callback_type].remove(self.__make_item(callback))

    def execute_callback(self, callback_enum: Enum, *arugs, **kwargs) -> None:
        """
        Execute synchronizing callback functions with the given enum
        """
        callback_enum = self.__get_valid_enum(callback_enum)

        thread_id = get_ident
        for cb in self.callbacks[callback_enum]:
            if self.thread_safe and not cb[0] == thread_id:
                raise RuntimeError(BaseLocal.get_message(
                    LocalCode_Invalid_Thread_Safe_Call, thread_id, cb[0]))

            cb[1](*arugs, **kwargs)

    async def execute_callback_async(self, callback_enum: Union[Enum, str], *arugs, **kwargs) -> None:
        """
        Execute asynchronizing callback functions with the given enum
        """
        callback_enum = self.__get_valid_enum(callback_enum)

        thread_id = get_ident
        for cb in self.callbacks[callback_enum]:
            if self.thread_safe and not cb[0] == thread_id:
                raise RuntimeError(BaseLocal.get_message(
                    LocalCode_Invalid_Thread_Safe_Call, thread_id, cb[0]))

            if iscoroutinefunction(cb[1]):
                await cb[1](*arugs, **kwargs)
            else:
                raise TypeError(BaseLocal.get_message(
                    LocalCode_Not_ASYNC_FUNC, cb[1].__name__))

    def __make_item(self, callback: Callable) -> Tuple:
        return (get_ident, callback)

    def __get_valid_enum(self, callback_enum: Enum) -> Enum:
        callback_type = self.callback_type_cls(callback_enum)

        if not callback_type in self.callbacks:
            raise TypeError(BaseLocal.get_message(
                LocalCode_Not_Valid_Enum, callback_enum, self.callback_type_cls))

        return callback_type



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: August 27th 2020 12:27:01 pm
'''

Localization_File = 'local.csv'

# base localization code
LocalCode_No_Code = 0                                   # args: (str, str)
LocalCode_Duplicated_Code = 1                           # args: (str, str)
LocalCode_Local_Pack_Parsing_Error = 2                  # args: (str)
LocalCode_Message_Format_Invalid = 3                    # args: (Tuple)

# asynccontextmanager
LocalCode_Not_Async_Gen = 10                            # args: ()
LocalCode_Not_Yield = 11                                # args: ()
LocalCode_Not_Stop = 12                                 # args: ()
LocalCode_Not_Stop_After_Throw = 13                     # args: ()

# callbacks
LocalCode_Not_Valid_Enum = 20                           # args: (Union[Enum, str], Enum)
LocalCode_Not_ASYNC_FUNC = 21                           # args: (Callable)
LocalCode_Invalid_Thread_Safe_Call = 22                 # args: (int, int)

# dynamic class enum
LocalCode_Must_Be_3Str_Tuple = 30                       # args: ()
LocalCode_Must_Be_Class = 31                            # args: (Type)
LocalCode_Must_Be_Function = 32                         # args: (Callable)
LocalCode_Is_Not_Subclass = 33

# hierarchy element
LocalCode_Not_HierarchyElementMeta_Subclass = 40        # args: (str)
LocalCode_No_Parameters = 41                            # args: (type)
LocalCode_Parameters_No_Key = 42                        # args: (type, str)
